from itertools import chain
from collections import defaultdict
import numpy as np
import offsetbasedgraph as obg
import itertools
from .ioclass import CollectionIO

class Areas(object):
    pass


class BinaryContinousAreas(Areas):
    def __init__(self, graph):
        self.graph = graph
        self.full_areas = {}
        self.starts = defaultdict(int)
        self.internal_intervals = {}

    def __str__(self):
        full_str = "Full: %s" % ",".join(
            str(node_id) for node_id in self.full_areas)
        start_str = "Start: %s" % ", ".join(
            "(%s,%s)" % (node_id, idx) for node_id, idx in self.starts.items())
        internal_str = "Internsals: %s" % self.internal_intervals
        return "\n".join((full_str, start_str, internal_str))

    __repr__ = __str__

    def __eq__(self, other):
        if self.full_areas != other.full_areas:
            return False
        if self.starts != other.starts:
            return False
        return self.internal_intervals == other.internal_intervals

    def add_full(self, node_id):
        self.full_areas[abs(node_id)] = 1

    def add_start(self, node_id, idx):
        assert idx > 0
        self.starts[node_id] = max(idx, self.starts[node_id])

    def _is_end_position(self, node, offset):
        if offset < self.graph.node_size(node):
            return True
        graph = self.graph
        for in_node in itertools.chain(graph.adj_list[node], graph.reverse_adj_list[node]):
            if in_node in self.full_areas or in_node in self.starts or -in_node in self.full_areas:
                return False

        return True

    def _is_start_position(self, node, offset):
        if offset > 0:
            return True

        graph = self.graph
        for in_node in itertools.chain(graph.adj_list[-node], graph.reverse_adj_list[-node]):
            if in_node in self.full_areas or in_node in self.starts or -in_node in self.full_areas:
                return False

        return True

    def merge_with_other(self, other):
        for full in other.full_areas.keys():
            self.add_full(full)
        for node, start in other.starts.items():
            self.add_start(node, start)
        for node, startend in other.internal_intervals.items():
            self.add_internal(node, startend[0], startend[1])

    def add_internal(self, node_id, start, end):
        assert start != end
        self.internal_intervals[node_id] = [start, end]

    def add(self, node_id, start, end):
        assert start != end
        node_size = self.graph.node_size(node_id)
        if start == 0 and end == node_size:
            self.add_full(node_id)
        elif start == 0:
            self.add_start(node_id, end)
        elif end == node_size:
            self.add_start(-node_id, node_size-start)
        else:
            if node_id < 0:
                start, end = node_size-end, node_size-start
                node_id = -node_id
            self.add_internal(node_id, start, end)

    def join_starts_and_ends(self):
        negative_node_ids = [node_id for node_id in self.starts if
                             node_id < 0]
        for node_id in negative_node_ids:
            if -node_id not in self.starts:
                continue
            if self.starts[node_id] + self.starts[-node_id] < self.graph.node_size(-node_id):
                continue
            self.full_areas[-node_id] = 1
            del self.starts[node_id]
            del self.starts[-node_id]

    def sanitize(self):
        for node_id in self.full_areas:
            if node_id in self.starts:
                del self.starts[node_id]
            if -node_id in self.starts:
                del self.starts[-node_id]
        self.join_starts_and_ends()

    def _add_abs_internal(self, node_id, start, end):
        if node_id < 0:
            node_size = self.graph.node_size(node_id)
            rev_start = node_size-end
            rev_end = node_size-start
            self.add_internal(abs(node_id), rev_start, rev_end)
        else:
            self.add_internal(node_id, start, end)

    def filled_interval(self, interval, fill_pos=0, fill_neg=0):
        start = max(interval.start_position.offset-fill_neg, 0)
        end_size = self.graph.node_size(interval.end_position.region_path_id)
        end = min(interval.end_position.offset+fill_pos,
                  end_size)
        neg_remain = max(0, fill_neg-interval.start_position.offset)
        pos_remain = max(0, fill_pos-(end_size-interval.end_position.offset))

        if len(interval.region_paths) == 1:
            # Make sure it touches edge if possible
            rp = interval.region_paths[0]
            if start == 0 and end == end_size:
                self.add_full(abs(rp))
            elif start == 0:
                self.add_start(rp, end)
            elif end == end_size:
                self.add_start(-rp, end-start)
            else:
                self._add_abs_internal(rp, start, end)
            return pos_remain, neg_remain
        start_rp = interval.region_paths[0]
        start_len = self.graph.node_size(start_rp) - start
        self.add_start(-start_rp, start_len)
        self.add_start(interval.end_position.region_path_id, end)
        for region_path in interval.region_paths[1:-1]:
            self.add_full(abs(region_path))
        return pos_remain, neg_remain

    def get_start_positions(self):
        start_positions = [
            obg.Position(-node_id, self.graph.node_size(node_id)-offset)
            for node_id, offset in self.starts.items()]
        node_ids = list(self.full_areas.keys()) + [
            -node_id for node_id in self.full_areas]
        previous_nodes_list = [
            self.graph.reverse_adj_list[-node_id]
            for node_id in node_ids]

        def filter_my_nodes(node_list):
            return [node_id for node_id in node_list if
                    abs(node_id) in self.full_areas or node_id
                    in self.starts]
        previous_nodes_list = [filter_my_nodes(prev_nodes) for
                               prev_nodes in previous_nodes_list]
        full_starts = [obg.Position(node_id, 0) for node_id, prev_nodes in
                       zip(node_ids, previous_nodes_list)
                       if not prev_nodes]

        return start_positions + full_starts

    def get_node_ids(self):
        return chain(self.full_areas.keys(), self.starts.keys(),
                     self.internal_intervals.keys())

    @classmethod
    def from_old_areas(cls, areas):
        binary_connected_areas = cls(areas.graph)
        for node_id in areas.areas:
            starts = areas.get_starts(node_id)
            ends = areas.get_ends(node_id)
            for start, end in zip(starts, ends):
                binary_connected_areas.add(node_id, start, end)

        return binary_connected_areas

    def to_file_line(self):
        fulls = ",".join(str(node_id) for node_id in self.full_areas)
        starts = ",".join("%s:%s" % (node_id, offset)
                          for node_id, offset in self.starts.items())
        internals = ",".join("%s:%s-%s" % (node_id, v[0], v[1])
                             for node_id, v in self.internal_intervals.items())
        return "(%s)\t(%s)\t(%s)\n" % (fulls, starts, internals)

    @classmethod
    def from_file_line(cls, line, graph):
        obj = cls(graph)
        fulls, starts, internals = [v[1:-1] for v in line.split()]
        if fulls:
            obj.full_areas = {int(full): 1 for full in fulls.split(",")}
        if starts:
            obj.starts = {int(p.split(":")[0]): int(p.split(":")[1]) for
                          p in starts.split(",")}
        if internals:
            k, v = internals.split(":")
            obj.internal_intervals = {int(k): [int(v.split("-")[0]),
                                               int(v.split("-")[1])]}
        return obj


class BCACollection(CollectionIO):
    _obj_type = BinaryContinousAreas


class ValuedAreas(Areas):
    def __init__(self, graph):
        self.graph = graph
        self.full_areas = defaultdict(int)
        self.starts = defaultdict(list)
        self.internal_starts = defaultdict(list)
        self.internal_ends = defaultdict(list)

    def has_anything_on_node(self, node_id):
        if node_id in self.full_areas:
            return True
        if node_id in self.starts:
            return True
        if node_id in self.internal_ends:
            return True
        if node_id in self.internal_starts:
            return True
        return False

    def add_binary_areas(self, areas, touched_nodes=None):
        for node_id in areas.full_areas:
            self.full_areas[node_id] += 1
            touched_nodes.add(abs(node_id))
        for node_id, internal_intervals in areas.internal_intervals.items():
            self.internal_starts[node_id].append(internal_intervals[0])
            self.internal_ends[node_id].append(internal_intervals[1])
            touched_nodes.add(abs(node_id))
        for node_id, start in areas.starts.items():
            self.starts[node_id].append(start)
            touched_nodes.add(abs(node_id))

    def get_starts_array(self, node_id, node_size=None):
        if node_size is None:
            node_size = self.graph.node_size(node_id)
        """
        length_starts_full = len(self.starts[node_id])+self.full_areas[node_id]
        length_internal = len(self.internal_starts[node_id])
        length_starts = len(self.starts[-node_id])
        out = np.zeros(length_starts_full + length_internal + length_starts)
        b = length_starts_full+length_internal
        out[length_starts_full:b] = self.internal_starts[node_id]
        out[b:] = self.starts[-node_id]
        out[b:] *= -1
        out[b:] += node_size
        return out
        """

        starts = [0]*(len(self.starts[node_id])+self.full_areas[node_id])
        starts.extend(self.internal_starts[node_id])
        starts.extend([node_size-start for start in self.starts[-node_id]])
        return np.array(starts, dtype="int")

    def get_ends_array(self, node_id, node_size=None):
        if node_size is None:
            node_size = self.graph.node_size(node_id)

        ends = self.starts[node_id][:]
        ends.extend(self.internal_ends[node_id])
        ends.extend([node_size]*(
            len(self.starts[-node_id])+self.full_areas[node_id]))
        return np.array(ends, dtype="int")



# Time-stamp: <2015-03-10 15:38:17 Tao Liu>

"""Module Description
Copyright (c) 2008,2009 Yong Zhang, Tao Liu <taoliu@jimmy.harvard.edu>
Copyright (c) 2010,2011 Tao Liu <taoliu@jimmy.harvard.edu>
This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).
@status:  experimental
@version: $Revision$
@author:  Yong Zhang, Tao Liu
@contact: taoliu@jimmy.harvard.edu
"""
import numpy as np
import logging
from collections import defaultdict

def median(nums):
    """Calculate Median.
    Parameters:
    nums:  list of numbers
    Return Value:
    median value
    """
    p = sorted(nums)
    l = len(p)
    if l % 2 == 0:
        return (p[l//2]+p[l//2-1])//2
    else:
        return p[l//2]


class NotEnoughPairsException(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Treatment:
    def __init__(self, dicts):
        assert isinstance(dicts, dict)
        assert "+" in dicts and "-" in dicts and len(dicts) == 2

        self._chroms = list(set(list(dicts["+"].keys()) + list(dicts["-"].keys())))
        logging.info("Treatment init")
        logging.info("Chroms: %s" % self._chroms)
        self._pos_dict = {key: np.array(dicts["+"][key], dtype="int")
                          for key in self._chroms}
        self._neg_dict = {key: np.array(dicts["-"][key], dtype="int")
                          for key in self._chroms}
        for val in self._pos_dict.values():
            val.sort()
        for val in self._neg_dict.values():
            val.sort()
        self.total = sum(val.size for val in self._pos_dict.values())
        self.total += sum(val.size for val in self._neg_dict.values())

    def get_chr_names(self):
        return self._chroms
        return ["chr%s" % i for i in range(1, 10)]

    def get_locations_by_chr(self, chrom):
        return self._pos_dict[chrom], self._neg_dict[chrom]

        n_peaks = 10000
        a = np.array([1, 2, 3, 4, 5], dtype="int")
        pos_tags = np.empty(5*n_peaks, dtype="int")
        for i in range(n_peaks):
            pos_tags[(i*5):((i+1)*5)] = i*1000+a
        return pos_tags, pos_tags+50

    @classmethod
    def from_bed_file(cls, file_name):
        dicts = {"+": defaultdict(list),
                 "-": defaultdict(list)}
        with open(file_name) as f:
            for line in f:
                if line.startswith("#"):
                    continue
                parts = line.split()
                chr_name = parts[0]
                start = int(parts[1])
                end = int(parts[2])
                strand = parts[5]
                pos = start if strand == "+" else end
                dicts[strand][chr_name].append(pos)

        return cls(dicts)


class Opt:
    def __init__(self):
        self.umfold = 50
        self.lmfold = 5
        self.bw = 300
        self.info = logging.info
        self.debug = logging.debug
        self.warn = logging.warning
        self.gsize = 300000000


class PeakModel:
    """Peak Model class.
    """

    def __init__(self, opt, treatment, max_pairnum=500):
        self.treatment = treatment
        self.gz = opt.gsize
        self.umfold = opt.umfold
        self.lmfold = opt.lmfold
        self.tag_expansion_size = 10
        self.bw = opt.bw
        self.info = opt.info
        self.debug = opt.debug
        self.warn = opt.warn
        self.error = opt.warn
        self.max_pairnum = max_pairnum
        self.build()

    def build(self):
        """Build the model.
        prepare self.d, self.scan_window, self.plus_line,
        self.minus_line and self.shifted_line to use.
        """
        self.peaksize = 2*self.bw
        self.min_tags = int(round(
            float(self.treatment.total)*self.lmfold*self.peaksize/self.gz/2))
        # mininum unique hits on single strand
        self.max_tags = int(round(
            float(self.treatment.total)*self.umfold*self.peaksize/self.gz/2))
        # maximum unique hits on single strand
        # use treatment data to build model
        self.info("#2 looking for paired plus/minus strand peaks...")
        paired_peakpos = self.__paired_peaks()
        # select up to 1000 pairs of peaks to build model
        num_paired_peakpos = 0
        num_paired_peakpos_remained = self.max_pairnum
        num_paired_peakpos_picked = 0
        # select only num_paired_peakpos_remained pairs.
        for c in paired_peakpos.keys():
            num_paired_peakpos += len(paired_peakpos[c])
        # TL: Now I want to use everything

        num_paired_peakpos_picked = num_paired_peakpos

        self.info("#2 number of paired peaks: %d" % (num_paired_peakpos))
        if num_paired_peakpos < 100:
            logging.warning("Too few paired peaks (%d) to get good shift estimate" % num_paired_peakpos)
            raise NotEnoughPairsException("No enough pairs to build model")
        elif num_paired_peakpos < self.max_pairnum:
            logging.warning("Few paired peaks (%d) than %d! Model may not be build well! Lower your MFOLD parameter may erase this warning. Now I will use %d pairs to build model!" % (num_paired_peakpos, self.max_pairnum,num_paired_peakpos_picked))
        self.__paired_peak_model(paired_peakpos)

    def __paired_peak_model(self, paired_peakpos):
        """Use paired peak positions and treatment
        tag positions to build the model.
        Modify self.(d, model_shift size and
        scan_window size. and extra, plus_line,
        minus_line and shifted_line for plotting).
        """
        window_size = 1+2*self.peaksize+self.tag_expansion_size
        self.plus_line = np.zeros(window_size, dtype="int32")
        self.minus_line = np.zeros(window_size, dtype="int32")
        plus_start = np.zeros(window_size, dtype="int32")
        plus_end = np.zeros(window_size, dtype="int32")
        minus_start = np.zeros(window_size, dtype="int32")
        minus_end = np.zeros(window_size, dtype="int32")
        self.info("start model_add_line...")
        chroms = paired_peakpos.keys()
        for chrom in chroms:
            paired_peakpos_chrom = paired_peakpos[chrom]
            (tags_plus, tags_minus) = self.treatment.get_locations_by_chr(
                chrom)
            self.__model_add_line(paired_peakpos_chrom, tags_plus,
                                  plus_start, plus_end)
            self.__model_add_line(paired_peakpos_chrom, tags_minus,
                                  minus_start, minus_end)

        self.__count(plus_start, plus_end, self.plus_line)
        self.__count(minus_start, minus_end, self.minus_line)

        self.info("start X-correlation...")
        # Now I use cross-correlation to find the best d
        plus_line = self.plus_line
        minus_line = self.minus_line

        # normalize first
        minus_data = (minus_line - minus_line.mean())/(minus_line.std()*len(minus_line))
        plus_data = (plus_line - plus_line.mean())/(plus_line.std()*len(plus_line))

        # cross-correlation
        ycorr = np.correlate(minus_data, plus_data, mode="full")[
            window_size-self.peaksize:window_size+self.peaksize]
        xcorr = np.linspace(len(ycorr)//2*-1, len(ycorr)//2, num=len(ycorr))

        # smooth correlation values to get rid of local maximums from small fluctuations. 
        ycorr = smooth(ycorr, window="flat")

        # all local maximums could be alternative ds.
        i_l_max = np.r_[
            False,
            ycorr[1:] > ycorr[:-1]] & np.r_[ycorr[:-1] > ycorr[1:],
                                            False]
        tmp_cor_alternative_d = ycorr[i_l_max]
        tmp_alternative_d = xcorr[i_l_max]
        cor_alternative_d = tmp_cor_alternative_d[tmp_alternative_d > 0]
        self.alternative_d = [int(val) for val in
                              tmp_alternative_d[tmp_alternative_d > 0]]
        # best cross-correlation point
        self.d = xcorr[np.where(ycorr == max(cor_alternative_d))[0][0]]
        # get rid of the last local maximum if it's at the right end of curve.
        assert len(self.alternative_d) > 0, "No proper d can be found! Tweak --mfold?"

        self.ycorr = ycorr
        self.xcorr = xcorr

        self.scan_window = max(self.d, self.tag_expansion_size)*2
        return True

    def __model_add_line(self, pos1, pos2, start, end):
        """Project each pos in pos2 which is included in
        [pos1-self.peaksize, pos1+self.peaksize] to the line.
        pos1: paired centers -- array.array
        pos2: tags of certain strand -- a numpy.array object
        line: numpy array object where we pileup tags
        """
        psize_adjusted1 = self.peaksize + self.tag_expansion_size // 2

        i1 = 0                  # index for pos1
        i2 = 0                  # index for pos2
        i2_prev = 0             # index for pos2 in previous pos1
        i1_max = len(pos1)
        i2_max = pos2.shape[0]
        flag_find_overlap = False

        max_index = start.shape[0] - 1

        while i1 < i1_max and i2 < i2_max:
            p1 = pos1[i1]
            p2 = pos2[i2]
            if p1-psize_adjusted1 > p2:
                i2 += 1
            elif p1 + psize_adjusted1 < p2:
                i1 += 1
                i2 = i2_prev    # search minus peaks from previous index
                flag_find_overlap = False
            else:               # overlap!
                if not flag_find_overlap:
                    flag_find_overlap = True
                    i2_prev = i2
                s = max(p2-p1+self.peaksize, 0)
                start[s] += 1
                e = min(p2+self.tag_expansion_size-p1+self.peaksize,
                        max_index)
                end[e] -= 1
                i2 += 1
        return

    def __model_add_line_vanilla(self, pos1, pos2, start, end):
        """Project each pos in pos2 which is included in
        [pos1-self.peaksize, pos1+self.peaksize] to the line.
        pos1: paired centers -- array.array
        pos2: tags of certain strand -- a numpy.array object
        line: numpy array object where we pileup tags
        """
        i1 = 0                  # index for pos1
        i2 = 0                  # index for pos2
        i2_prev = 0             # index for pos2 in previous pos1
        i1_max = len(pos1)
        i2_max = pos2.shape[0]
        flag_find_overlap = False

        max_index = start.shape[0] - 1
        psize_adjusted1 = self.peaksize + self.tag_expansion_size // 2

        while i1 < i1_max and i2 < i2_max:
            p1 = pos1[i1]
            p2 = pos2[i2]
            if p1-psize_adjusted1 > p2:
                i2 += 1
            elif p1+psize_adjusted1 < p2:
                i1 += 1
                i2 = i2_prev    # search minus peaks from previous index
                flag_find_overlap = False
            else:               # overlap!
                if not flag_find_overlap:
                    flag_find_overlap = True
                    i2_prev = i2
                s = max(p2-self.tag_expansion_size//2-p1+psize_adjusted1, 0)
                start[s] += 1
                e = min(p2+self.tag_expansion_size//2-p1+psize_adjusted1,
                        max_index)
                end[e] -= 1
                i2 += 1
        return

    def __count(self, start, end, line):
        pileup = 0
        for i in range(line.shape[0]):
            pileup += start[i] + end[i]
            line[i] = pileup
        return

    def __paired_peaks(self):
        """Call paired peaks from fwtrackI object.
        Return paired peaks center positions.
        """
        chrs = self.treatment.get_chr_names()
        chrs.sort()
        paired_peaks_pos = {}
        for i in range(len(chrs)):
            chrom = chrs[i]
            plus_tags, minus_tags = self.treatment.get_locations_by_chr(chrom)
            plus_peaksinfo = self.__naive_find_peaks(plus_tags, 1)
            minus_peaksinfo = self.__naive_find_peaks(minus_tags, 0)
            if not plus_peaksinfo or not minus_peaksinfo:
                continue
            paired_peaks_pos[chrom] = self.__find_pair_center(
                plus_peaksinfo, minus_peaksinfo)
        return paired_peaks_pos

    def __find_pair_center(self, pluspeaks, minuspeaks):
        ip = 0                  # index for plus peaks
        im = 0                  # index for minus peaks
        im_prev = 0             # index for minus peaks in previous plus peak
        pair_centers = []
        ip_max = len(pluspeaks)
        im_max = len(minuspeaks)
        flag_find_overlap = False
        while ip < ip_max and im < im_max:
            (pp, pn) = pluspeaks[ip]  # for (peakposition, tagnumber in peak)
            (mp, mn) = minuspeaks[im]
            if pp-self.peaksize > mp:  # move minus
                im += 1
            elif pp+self.peaksize < mp:  # move plus
                ip += 1
                im = im_prev    # search minus peaks from previous index
                flag_find_overlap = False
            else:               # overlap!
                if not flag_find_overlap:
                    flag_find_overlap = True
                    im_prev = im  # only the first index is recorded
                # number tags in plus and minus peak region are comparable...
                if float(pn)/mn < 2 and float(pn)/mn > 0.5:
                    if pp < mp:
                        pair_centers.append((pp+mp)//2)
                im += 1
        return pair_centers

    def __naive_find_peaks(self, taglist, plus_strand=1):
        """Naively call peaks based on tags counting.
        if plus_strand == 0, call peak on minus strand.
        Return peak positions and the tag number in peak
        region by a tuple list [(pos,num)].
        """
        peak_info = []    # store peak pos in every peak region and
                          # unique tag number in every peak region
        if taglist.shape[0] < 2:
            return peak_info
        pos = taglist[0]
        # current_tag_list = [pos]
        cur_start = 0
        for i in range(1, len(taglist)):
            pos = taglist[i]
            # call peak in current_tag_list when the region is long enough
            if (pos - taglist[cur_start] + 1) > self.peaksize:
                # a peak will be called if tag number is ge min tags.
                if self.max_tags >= i-cur_start >= self.min_tags:
                    peak_info.append((self.__naive_peak_pos(
                        taglist[cur_start:i], plus_strand),
                                      i-cur_start))
                cur_start = i

            # current_tag_list.append(pos)   # add pos while 1. no
                                             # need to call peak;
                                             # 2. current_tag_list is []
        return peak_info

    def __get_horizon_line_sparse(self, pos_list, start, peak_length):
        positions = np.empty(pos_list.size*2)
        positions[:pos_list.size] = np.maximum(
            pos_list-start-self.tag_expansion_size//2,
            0)
        positions[pos_list.size:] = np.minimum(
            pos_list-start+self.tag_expansion_size//2,
            peak_length)
        args = np.argsort(positions)
        codes = 1-2*(args >= pos_list.size)

    def __get_horizon_line(self, pos_list):
        # pos_list = np.array(pos_list, dtype="int")
        peak_length = pos_list[-1]+1-pos_list[0] + self.tag_expansion_size
        # leftmost position of project line
        start = pos_list[0] - self.tag_expansion_size//2
        ss = pos_list-start-self.tag_expansion_size//2
        es = pos_list-start+self.tag_expansion_size//2

        # the line for tags to be projected
        horizon_line = np.zeros(peak_length, dtype="int32")
        horizon_line[ss] += 1
        horizon_line[es] -= 1
        horizon_line = np.cumsum(horizon_line)
        return horizon_line

    def __naive_peak_pos(self, pos_list, plus_strand):
        """Naively calculate the position of peak.
        plus_strand: 1, plus; 0, minus
        return the highest peak summit position.
        """
        peak_length = pos_list[-1]+1-pos_list[0] + self.tag_expansion_size
        start = pos_list[0] - self.tag_expansion_size//2
        horizon_line = self.__get_horizon_line(pos_list)
        top_pos = self.__find_top_pos(horizon_line, peak_length)
        return (top_pos[len(top_pos)//2]+start)


    def __find_top_pos(self, horizon_line, peak_length):
        m = np.max(horizon_line)
        return np.where(horizon_line == m)[0]



# smooth function from SciPy cookbook: http://www.scipy.org/Cookbook/SignalSmooth
def smooth(x, window_len=11, window='hanning'):
    """smooth the data using a window with requested size.
    
    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal 
    (with the window size) in both ends so that transient parts are minimized
    in the beginning and end part of the output signal.
    
    input:
        x: the input signal 
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.
    output:
        the smoothed signal
        
    example:
    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)
    
    see also: 
    
    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter
 
    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len < 3:
        return x

    if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError(
            "Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")

    s = np.r_[x[window_len-1:0:-1], x, x[-1:-window_len:-1]]
    if window == 'flat':  # moving average
        w = np.ones(window_len, 'd')
    else:
        w = eval('np.'+window+'(window_len)')

    y = np.convolve(w/w.sum(), s, mode='valid')
    return y[(window_len//2):-(window_len//2)]




import numpy as np


class LinearIntervalCollection(object):

    def __init__(self, starts, ends):
        self.starts = np.asanyarray(starts)
        self.ends = np.asanyarray(ends)
        self.n_intervals = self.starts.size

    def __str__(self):
        return str(self.starts) + "\n" + str(self.ends)

    def extend_mid(self, extension_size):
        extended_starts = (self.starts + self.ends)/2 - extension_size
        extended_ends = (self.starts + self.ends)/2 + extension_size
        return self.__class__(extended_starts, extended_ends)

    def extend(self, extension_size):
        extended_starts = self.starts - extension_size
        extended_ends = self.starts + extension_size
        return self.__class__(extended_starts, extended_ends)

    def extend_np(self, extension_size):
        return np.add.outer(np.array([-extension_size, extension_size]),
                            self.starts)

    def n_basepairs_covered(self):
        return np.sum(np.abs(self.ends - self.starts))



#!/usr/bin/python3
import argparse
import logging
import sys
import matplotlib as mpl
mpl.use('Agg')

import offsetbasedgraph as obg
from graph_peak_caller.peakcollection import Peak, PeakCollection

from graph_peak_caller.callpeaks_interface import \
    run_callpeaks_interface, run_callpeaks_whole_genome,\
    run_callpeaks_whole_genome_from_p_values

from graph_peak_caller.analysis_interface import analyse_peaks_whole_genome,\
    analyse_peaks, intervals_to_fasta, linear_peaks_to_fasta,\
    analyse_manually_classified_peaks, differential_expression,\
    plot_motif_enrichment

from graph_peak_caller.preprocess_interface import shift_estimation,\
    count_unique_reads_interface, create_ob_graph,\
    create_linear_map_interface,\
    split_vg_json_reads_into_chromosomes


logging.basicConfig(
    stream=sys.stdout, level=logging.WARNING,
    format="%(asctime)s, %(levelname)s: %(message)s")


def main():
    create_argument_parser()

def version(args):
    print("Graph Peak Caller v1.0.4")

def concatenate_sequence_files(args):
    chromosomes = args.chromosomes.split(",")
    out_file_name = args.out_file_name

    all_fasta_entries = []
    for chromosome in chromosomes:
        print("Processing chromosome %s" % chromosome)
        fasta_file = open(chromosome + "_sequences.fasta")
        for line in fasta_file:
            if line.startswith(">"):
                all_fasta_entries.append([line, None])
            else:
                # This is sequence, add to prev entry
                all_fasta_entries[-1][1] = line

    peaks = []
    for fasta_entry in all_fasta_entries:
        header = fasta_entry[0].rstrip()
        sequence = fasta_entry[1].rstrip()

        interval_json = "{%s" % header.split(" {")[1]
        interval = Peak.from_file_line(interval_json)
        interval.sequence = sequence
        peaks.append(interval)

    peaks = sorted(peaks, key=lambda s: -s.score)
    out_fasta = open(out_file_name, "w")
    i = 0
    for peak in peaks:
        out_fasta.writelines([">peak%d %s\n" % (i, peak.to_file_line())])
        out_fasta.writelines(["%s\n" % peak.sequence])
        i += 1
    out_fasta.close()

    logging.info("Wrote all peaks in sorted order to %s" % out_file_name)


def find_linear_path(args):
    from graph_peak_caller.util import create_linear_path
    import pyvg
    graph = obg.GraphWithReversals.from_numpy_file(args.ob_graph_file_name)
    vg_graph = pyvg.Graph.from_file(args.vg_json_graph_file_name)
    linear_path = create_linear_path(graph, vg_graph,
                                     path_name=args.linear_path_name,
                                     write_to_file=None)
    linear_path.to_file(args.out_file_name)
    logging.info("Wrote to file %s" % args.out_file_name)


def move_linear_reads_to_graph(args):
    chromosomes = args.chromosomes.split(",")
    chrom_lookup = set(chromosomes)
    graphs = {}
    out_files = {}
    linear_paths = {}
    for chrom in chromosomes:
        linear_paths[chrom] = obg.NumpyIndexedInterval.from_file(
            args.data_dir + "/" + chrom + "_linear_pathv2.interval"
        )
        graphs[chrom] = obg.Graph.from_numpy_file(args.data_dir + "/" + chrom + ".nobg")
        out_files[chrom] = open(args.out_files_base_name + "_" + chrom + ".intervalcollection", "w")

    bedfile = open(args.bed_file_name, "r")
    i = 0
    for line in bedfile:
        if i % 100000 == 0:
            logging.info("%d reads processed" % i)
        i += 1
        line = line.split()
        is_reverse = line[5] == "-"
        chrom = line[0].replace("chr", "")
        if chrom not in chrom_lookup:
            continue
        start = int(line[1])
        end = int(line[2])
        graph_interval = linear_paths[chrom].get_exact_subinterval(start, end)
        graph_interval.graph = graphs[chrom]

        if is_reverse:
            graph_interval = graph_interval.get_reverse()
            assert graph_interval.region_paths[0] < 0
        out_files[chrom].writelines(["%s\n" % graph_interval.to_file_line()])

    for chrom in chromosomes:
        out_files[chrom].close()


def peaks_to_linear(args):
    # Get approximate linear position of peaks using a linear path
    linear_path = obg.NumpyIndexedInterval.from_file(args.linear_path_file_name)
    peaks = PeakCollection.from_file(args.peaks_file_name, text_file=True)
    linear_peaks = peaks.to_approx_linear_peaks(linear_path, args.chromosome)
    linear_peaks.to_bed_file(args.out_file_name)


interface = \
{
    'callpeaks':
        {
            'help': 'Callpeaks',
            'arguments':
                [
                    ('graph_file_name', ""),
                    ('vg_graph_file_name', "Graph file name (.vg). Used for fetching sequences. Can be set to 'None'."),
                    ('linear_map_base_name', "Set to base name of linear map."),
                    ('sample_reads_file_name', 'File name to a vg JSON file or intervalcollection file.'),
                    ('control_reads_file_name', 'File name to a vg JSON file or intervalcollection file. Set to the same as sample if no control.'),
                    ('with_control', 'True/False. Set to False if control is the same as sample.'),
                    ('out_base_name', 'eg experiment1_. Will be preprended to all output files.'),
                    ('fragment_length', 'The fragment length used in this ChIP-seq experiment. If unknown, set to an '
                                        'arbitrary number, e.g. 200. However, for good results, this number should be accurate.'),
                    ('read_length', 'The read length.')
                ],
            'method': run_callpeaks_interface
        },
    'callpeaks_whole_genome':
        {
            'help': 'Callpeaks on whole genome, using one graph for each chromosome',
            'arguments':
                [
                    ('chromosomes', 'Comma-separated list of chromosomes to use, e.g. 1,2,X,8,Y'),
                    ('graphs_location', 'Will use the graphs *_[chromosome]'),
                    ('vg_graphs_location', ''),
                    ('linear_maps_location', ''),
                    ('sample_reads_base_name', 'Will use files *_[chromosome].json where * is the base name'),
                    ('control_reads_base_name', 'Will use files *_[chromosome].json where * is the base name'),
                    ('out_base_name', 'eg experiment1_'),
                    ('with_control', 'True/False'),
                    ('fragment_length', ''),
                    ('read_length', ''),
                    ('stop_after_p_values', 'True/False - whether to only run until p-value track is computed (before peak calling)'),
                    ('unique_reads', 'Number of unique reads. Found by calling count_unique_reads'),
                    ('genome_size', 'Number of base pairs covered by graphs in total (on a linear genome)')
                ],
            'method': run_callpeaks_whole_genome
        },
    'callpeaks_whole_genome_from_p_values':
        {
            'help': 'Callpeaks on whole genome from pvalues. Assumes pvalue files are already computed.',
            'arguments':
                [
                    ('chromosome', 'Specific chromosome to find peaks for.'),
                    ('graphs_location', 'Directory containing graphs.'),
                    ('out_base_name', 'eg experiment1_'),
                    ('with_control', 'True/False'),
                    ('fragment_length', ''),
                    ('read_length', '')
                ],
            'method': run_callpeaks_whole_genome_from_p_values
        },
    'estimate_shift':
        {
            'help': 'Estimate shift on whole genome',
            'arguments':
                [
                    ('chromosomes', 'Comma-separated list of chromosomes to use, e.g. 1,2,X,8,Y'),
                    ('ob_graphs_location', 'Location of graph files'),
                    ('sample_reads_base_name', 'Will use files [base_name][chromosome].json'),
                ],
            'method': shift_estimation
        },
    'intervals_to_fasta':
        {
            'help': 'Get sequences for intervals in interval file. Write to fasta',
            'arguments':
                [
                    ('vg_graph_file_name', ''),
                    ('intervals_file_name', ''),
                    ('out_file_name', '')
                ],
            'method': intervals_to_fasta
        },
    'linear_peaks_to_fasta':
        {
            'help': 'Converts a linear peaks file (eg. from macs) to fasta',
            'arguments':
                [
                    ('linear_reads_file_name', ''),
                    ('fasta_file', ''),
                    ('out_file_name', '')
                ],
            'method': linear_peaks_to_fasta
        },
    'create_ob_graph':
        {
            'help': 'Creates Offset Based Graph from a vg json graph file. Stores the resulting graph to a file.',
            'arguments':
                [
                    ('vg_json_file_name', 'Vg json file name (created by running vg view -Vj graph.vg > graph.json'),
                    ('out_file_name', 'E.g. graph.obg')
                ],
            'method': create_ob_graph
        },
    'create_linear_map':
        {
            'help': 'Creates a linear map using a vg snarls file and an ob graph',
            'arguments':
                [
                    ('obg_file_name', ''),
                    ('vg_snarls_file_name', ''),
                    ('out_file_base_name', ''),
                ],
            'method': create_linear_map_interface
        },
    'split_vg_json_reads_into_chromosomes':
        {
            'help': "Splits intervals from interval collection into one file for each chromsome."
                    "Requires node_range_[chrom ID].txt to exist for each chromosome.",
            'arguments':
                [
                    ('chromosomes', 'Comma-separated list of chromosomes to split reads into'),
                    ('vg_json_reads_file_name', ''),
                    ('range_files_base_name', 'Base name, e.g. dir, to range files')
                ],
            'method': split_vg_json_reads_into_chromosomes
        },
    'concatenate_sequence_files':
        {
            'help': "Merge multiple *_sequence.fasta files from the peak caller into one single sorted file.",
            'arguments':
                [
                    ('chromosomes', 'comma delimted, e.g 1,2,3, used to fetch files of type chr1_sequences.fasta, ...'),
                    ('out_file_name', '')
                ],
            'method': concatenate_sequence_files
        },
    'plot_motif_enrichment':
        {
            'help': "Plots motif enrichments using fimo. Requires fimo to be installed and in path.",
            'arguments':
                [
                    ('fasta1', ''),
                    ('fasta2', ''),
                    ('meme_motif_file', 'something.meme'),
                    ('out_figure_file_name', ''),
                    ('run_fimo', 'Set to True if fimo has not already been run.'),
                    ('plot_title', 'Title above plot')
                ],
            'method': plot_motif_enrichment
        },
    'analyse_peaks':
        {
            'help': 'Analyse linear peaks and graph peaks.',
            'arguments':
                [
                    ('ob_graph_file_name', ''),
                    ('vg_graph_file_name', ''),
                    ('linear_peaks_fasta_file_name', ''),
                    ('graph_peaks_fasta_file_name', ''),
                    ('linear_peaks_fimo_results_file', ''),
                    ('graph_peaks_fimo_results_file', ''),
                    ('linear_path_name', 'Ref of chromosome name. Used to find path in vg json file'),
                    ('graph_chromosome', 'None if not relevant'),
                    ('graph_start', 'Start pos in chromosome of graph.'),
                    ('graph_end', 'End pos in chromosome of graph. 0 if covering whole chromosome')
                ],
            'method': analyse_peaks
        },
    'analyse_peaks_whole_genome':
        {
            'help': "Analyse peak results from whole genome run",
            'arguments':
                [
                    ('chromosomes', 'Comma seaparated list of chromosomes to use'),
                    ('results_dir', 'Directory of result files (should contain fasta files from both linear and graph peak calling'),
                    ('graphs_dir', 'Dir containing obg graphs on form 1.nobg, 2.nobg, ... and vg graphs 1.json, 2.json, ...'),
                    ('out_file', 'Out file base name (file endings for different formats will be appended)')
                ],
            'method': analyse_peaks_whole_genome
        },
    'count_unique_reads':
        {
            'help': 'Count unique reads for whole genome set of read files.',
            'arguments':
                [
                    ('chromosomes', 'Comma-separated list of chromosomes to use, e.g. 1,2,X,8,Y'),
                    ('graphs_location', 'Will use the graphs *_[chromosome]'),
                    ('reads_base_name', 'Will use files *_[chromosome].json')
                ],
            'method': count_unique_reads_interface
        },
    'analyse_manually_classified_peaks':
        {
            'help': '',
            'arguments':
                [
                    ('chromosomes', 'Comma-separated list of chromosomes to use, e.g. 1,2,X,8,Y'),
                    ('graphs_location', 'Will use the graphs *_[chromosome]'),
                    ('reads_base_name', 'Will use files *_[chromosome].json'),
                    ('regions_file', ''),
                    ('manually_classified_peaks_file', '')
                ],
            'method': analyse_manually_classified_peaks
        },
    'find_linear_path':
        {
            'help': 'Finds lineat path through graph. Saves as indexed interval to file.',
            'arguments':
                [
                    ('vg_json_graph_file_name', ''),
                    ('ob_graph_file_name', ''),
                    ('linear_path_name', 'Name of path in the vg graph (typically ref or chromosome name'),
                    ('out_file_name', ''),
                ],
            'method': find_linear_path
        },
    'move_linear_reads_to_graph':
        {
            'help': 'Translate reads in sam file to reads on a graph. Writes to interval collection files.',
            'arguments':
                [
                    ('bed_file_name', ''),
                    ('chromosomes', 'Comma separated list of chromosomes to use'),
                    ('data_dir', 'Directory containing graphs and linear path files'),
                    ('out_files_base_name', '')
                ],
            'method': move_linear_reads_to_graph
        },
    'diffexpr':
        {
            'help': 'Find differentially expressed motif matches',
            'arguments':
                [
                    ('test_name', ''),
                    ('graph_name', ''),
                    ('fimo_file_name', ''),
                    ('vg_graph_name', '')
                ],
            'method': differential_expression
        },
    'peaks_to_linear':
        {
            'help': 'Converts graph peaks to linear peaks by using a linear path through the graph',
            'arguments':
                [
                    ('peaks_file_name', "Shuold be a JSON intervalcollection, e.g. one created by callpeaks."),
                    ('linear_path_file_name', "Name of linera path file"),
                    ('chromosome', 'Name of chromosome that will be used when writing the bed file'),
                    ('out_file_name', 'Out file name')
                ],
            'method': peaks_to_linear
        },
    'version':
        {
            'help': 'Prints the current version',
            'arguments': [],
            'method': version
        }
}


def create_argument_parser():
    # Create parser
    parser = argparse.ArgumentParser(
        description='Graph peak caller')
    subparsers = parser.add_subparsers(help='Subcommands')

    for command in interface:
        example = ""
        if "example_run" in interface[command]:
            example = "\nExample: " + interface[command]["example_run"]

        subparser = subparsers.add_parser(
            command,
            help=interface[command]["help"] + example)
        for argument, help in interface[command]["arguments"]:
            subparser.add_argument(argument, help=help)
        subparser.set_defaults(func=interface[command]["method"])

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()
    if hasattr(args, 'func'):
        args.func(args)
    else:
        parser.help()

if __name__ == "__main__":
    create_argument_parser()



import logging
from pyvg.conversion import vg_json_file_to_interval_collection,\
    json_file_to_obg_numpy_graph
import offsetbasedgraph as obg

from graph_peak_caller.util import create_linear_map
from graph_peak_caller.multiplegraphscallpeaks import MultipleGraphsCallpeaks
from graph_peak_caller.shift_estimation_multigraph import \
    MultiGraphShiftEstimator


def shift_estimation(args):
    """
    from graph_peak_caller.shiftestimation import Treatment, Opt, PeakModel
    treatment = Treatment.from_bed_file("linear_bed.bed")
    opt = Opt()
    opt.gsize = 48172484
    model = PeakModel(opt, treatment)
    print(model.d)
    return
    """
    chromosomes = args.chromosomes.split(",")
    graphs = [args.ob_graphs_location + chrom for chrom in chromosomes]
    logging.info("Will try to use graphs %s" % graphs)
    sample_file_names = [args.sample_reads_base_name + chrom + ".json"
                         for chrom in chromosomes]
    logging.info("Will use reads from %s" % sample_file_names)

    estimator = MultiGraphShiftEstimator.from_files(
        chromosomes, graphs, sample_file_names)

    estimator.to_linear_bed_file("linear_bed.bed", read_length=36)

    d = estimator.get_estimates()
    print("Shift: %d" % d)


def count_unique_reads_interface(args):
    chromosomes = args.chromosomes.split(",")
    graph_file_names = [args.graphs_location + chrom for chrom in chromosomes]
    reads_file_names = [args.reads_base_name + chrom + ".json"
                        for chrom in chromosomes]

    count_unique_reads(chromosomes, graph_file_names,
                       reads_file_names)


def count_unique_reads(chromosomes, graph_file_names, reads_file_names):
    graphs = (obg.GraphWithReversals.from_numpy_file(f)
              for f in graph_file_names)
    reads = (vg_json_file_to_interval_collection(f, graph)
             for f, graph in zip(reads_file_names, graphs))

    unique_reads = MultipleGraphsCallpeaks.count_number_of_unique_reads(reads)
    print(unique_reads)


def create_ob_graph(args):
    logging.info("Creating obgraph")
    ob_graph = json_file_to_obg_numpy_graph(args.vg_json_file_name, 0)

    logging.info("Writing ob graph to file")
    ob_graph.to_numpy_file(args.out_file_name)


def create_linear_map_interface(args):
    logging.info("Reading ob graph from file")
    ob_graph = obg.GraphWithReversals.from_numpy_file(args.obg_file_name)
    logging.info("Converting to dict format, allowing graph to "
                 "be changed by linear map process")
    ob_graph.convert_to_dict_backend()
    logging.info("Creating linear map")
    create_linear_map(ob_graph, args.vg_snarls_file_name,
                      args.out_file_base_name, copy_graph=False)


def split_vg_json_reads_into_chromosomes(args):
    reads_base_name = args.vg_json_reads_file_name.split(".")[0]
    logging.info("Will write reads to files %s_[chromosome].json",
                 reads_base_name)

    chromosomes = args.chromosomes.split(",")
    chromosome_limits = {}
    logging.info("Found the following chromosome ranges:")
    for chrom in chromosomes:
        start_end = open(
            args.range_files_base_name + "node_range_" + chrom + ".txt")
        start_end = start_end.read().split(":")
        start = int(start_end[0])
        end = int(start_end[1])
        chromosome_limits[chrom] = (start, end)
        logging.info("   Chr%s: %d-%d" % (chrom, start, end))

    out_files = {chrom: open(reads_base_name + "_" + chrom + ".json", "w")
                 for chrom in chromosomes}

    reads_file = open(args.vg_json_reads_file_name)
    i = 0
    import re
    regex = re.compile(r"node_id\": ([0-9]+)")

    def get_mapped_chrom(node):
        mapped_chrom = None
        for chrom in chromosomes:
            if node >= chromosome_limits[chrom][0] and node <= chromosome_limits[chrom][1]:
                mapped_chrom = chrom
                break
        return mapped_chrom

    n_without_node_id = 0
    for line in reads_file:
        if i % 100000 == 0:
            logging.info("Line #%d" % i)
        i += 1

        groups = regex.search(line)
        if groups is None:
            n_without_node_id += 1
            continue
        groups = groups.groups()
        if len(groups) > 0:
            node = int(groups[0])
            mapped_chrom = get_mapped_chrom(node)
            if mapped_chrom is None:
                n_without_node_id += 1
                continue
            out_files[mapped_chrom].writelines([line])
        else:
            print("No groups fond")

    for file in out_files.values():
        file.close()

    logging.info("Done. Found %d lines without node id or not matching into the given list of chromosomes" % n_without_node_id)



import numpy as np
import logging
from scipy.sparse import csr_matrix

from .segmentanalyzer import SegmentSplitter
from ..peakcollection import Peak
from .holecleaner import PosDividedLineGraph, SubGraph


class SparseMaxPaths:
    def __init__(self, sparse_values, graph, score_pileup):
        self._node_indexes = graph.node_indexes
        self._sparse_values = sparse_values
        self._graph = graph
        self._score_pileup = score_pileup
        self._segments = self.get_segments()
        self._node_ids = self.get_node_ids()
        self._analyzer = SegmentSplitter(self._segments, self._node_ids,
                                         graph.node_indexes)

    def _handle_internal(self, mask):
        ids = self._analyzer._internal_ids[:, 0][mask]
        positions = self._analyzer._internal_segments[mask]
        offsets = positions-self._node_indexes[ids-1, None]
        ids -= self._graph.min_node-1
        logging.info("Found %s internal peaks", len(ids))
        self.internal_paths = [Peak(offset[0], offset[1], [_id],
                                    graph=self._graph)
                               for offset, _id in zip(offsets, ids)]

        for path in self.internal_paths:
            path.info = (False, False)

    def get_segments(self):
        start_idx = 0
        if self._sparse_values.values[0] != 1:
            start_idx += 1
        indices = self._sparse_values.indices[start_idx:]
        if self._sparse_values.values[-1] == 1:
            indices = np.r_[indices, self._node_indexes[-1]]
        return indices.reshape(indices.size//2, 2)

    def get_node_ids(self):
        node_ids = np.empty_like(self._segments)
        node_ids[:, 0] = np.digitize(self._segments[:, 0], self._node_indexes)
        node_ids[:, 1] = np.digitize(self._segments[:, 1],
                                     self._node_indexes, True)
        return node_ids

    def run(self):
        self._analyzer.run()
        self._segments = self._analyzer.splitted_segments
        self.get_segment_scores()
        scored_segments = [np.vstack((self._analyzer._internal_ids[:, 0][mask],
                                      self.scores[mask]))
                           for mask in [self._analyzer.internal_mask,
                                        self._analyzer.start_mask,
                                        self._analyzer.end_mask,
                                        self._analyzer.full_mask]]

        self._handle_internal(self._analyzer.internal_mask)
        linegraph = PosDividedLineGraph(scored_segments[2],
                                        scored_segments[3],
                                        scored_segments[1],
                                        self._graph)

        paths, infos, subgraphs = linegraph.max_paths()
        converted = self._convert_paths(paths, infos)
        # for path, subgraph in zip(converted, subgraphs):
        #     # subgraph._node_ids += self._graph.min_node-1
        #     print(subgraph._node_ids)
        #     print(path.region_paths)
        #     assert all(rp in subgraph._node_ids for rp in path.region_paths)

        small_subgraphs = [
            SubGraph(path.region_paths, csr_matrix(([], ([], [])), shape=(1, 1)))
                           for path in self.internal_paths]
        return converted+self.internal_paths, subgraphs+small_subgraphs

    def _convert_paths(self, paths, infos):
        reverse_map = np.concatenate(
            [np.flatnonzero(mask) for mask in
             [self._analyzer.end_mask,
              self._analyzer.full_mask,
              self._analyzer.start_mask,
              ]])
        peaks = [self._convert_path(path, reverse_map)
                 for path in paths]
        for peak, info in zip(peaks, infos):
            peak.info = info
        return peaks

    def _convert_path(self, path, reverse_map):
        idxs = reverse_map[path]
        node_ids = self._analyzer._internal_ids[:, 0]
        node_ids = node_ids[idxs]
        start_offset = self._segments[idxs[0], 0 ] - self._node_indexes[node_ids[0]-1]
        end_offset = self._segments[idxs[-1], 1] - self._node_indexes[node_ids[-1]-1]
        return Peak(start_offset, end_offset, list(node_ids + self._graph.min_node-1), self._graph)

    def get_segment_scores(self):
        pileup_idxs = np.empty_like(self._segments)
        pileup_idxs[:, 0] = np.digitize(self._segments[:, 0],
                                        self._score_pileup.indices)
        pileup_idxs[:, 1] = np.digitize(self._segments[:, 1],
                                        self._score_pileup.indices, True)
        weighted_values = np.ediff1d(
            self._score_pileup.indices,
            self._score_pileup.track_size-self._score_pileup.indices[-1])*self._score_pileup.values
        pileup_cumsum = np.r_[0, np.cumsum(weighted_values)]
        base_scores = pileup_cumsum[pileup_idxs[:, 1]-1]-pileup_cumsum[
            pileup_idxs[:, 0]-1]
        # assert np.all(base_scores >= 0)
        diffs = self._segments-self._score_pileup.indices[pileup_idxs-1]
        # assert np.all(diffs >= 0)
        values = self._score_pileup.values[pileup_idxs-1]
        # assert np.all(values[:, 1] >= values[:, 0])
        val_diffs = diffs*values
        offsets = val_diffs[:, 1] - val_diffs[:, 0]
        self.scores = base_scores + offsets
        # assert np.all(self.scores >= 0)



import logging
import numpy as np
from pybedtools import BedTool
import pyvg
import offsetbasedgraph as obg
from offsetbasedgraph.graphtraverser import GraphTraverserUsingSequence

from .control.linearsnarls import LinearSnarlMap
from .control.snarls import SnarlGraphBuilder


class LinearRegion(object):
    def __init__(self, chromosome, start, end):
        self.chromosome = chromosome
        self.start = start
        self.end = end

LRC_REGION = LinearRegion("chr19", 54025634, 55084318)
MHC_REGION = LinearRegion("chr6", 28510119, 33480577)


def bed_intervals_to_graph(obg_graph, linear_path_interval,
                           bed_file_name, graph_start_offset):
    f = open(bed_file_name)
    intervals_on_graph = []
    for line in f:
        l = line.split()
        start = int(l[1])
        end = int(l[2])

        start = start - graph_start_offset
        end = end - graph_start_offset
        intervals_on_graph.append(linear_path_interval.get_subinterval(start, end))

    return intervals_on_graph


def fasta_sequence_to_linear_path_through_graph(
        linear_sequence_fasta_file, sequence_retriever, ob_graph, start_node):
    search_sequence = open(linear_sequence_fasta_file).read()
    print("Length of search sequence: %d" % len(search_sequence))
    traverser = GraphTraverserUsingSequence(
        ob_graph, search_sequence, sequence_retriever)
    traverser.search_from_node(start_node)
    linear_path_interval = traverser.get_interval_found()
    return linear_path_interval


def get_linear_paths_in_graph(ob_graph, vg_graph, write_to_file_name=None):
    assert ob_graph is not None
    intervals = {}
    for path in vg_graph.paths:
        obg_interval = path.to_obg(ob_graph=ob_graph)
        obg_interval.name = path.name
        print("Path name: %s" % path.name)
        intervals[obg_interval.name] = obg_interval

    if write_to_file_name is not None:
        logging.info("Writing linear path to %s" % write_to_file_name)
        collection = obg.IntervalCollection(intervals.values())
        collection.to_file(write_to_file_name, text_file=True)

    return intervals


def sparse_maximum(indices1, values1, indices2, values2, genome_size):

    if indices1[0] != 0:
        np.insert(indices1, 0, 0)
        np.insert(values1, 0, 0)

    if indices2[0] != 0:
        np.insert(indices2, 0, 0)
        np.insert(values2, 0, 0)

    indices1 = np.insert(indices1, len(indices1), genome_size)
    indices2 = np.insert(indices2, len(indices2), genome_size)

    a = indices1[:-1] * 2
    b = indices2[:-1] * 2 + 1
    all_idxs = np.concatenate([a, b])
    all_idxs.sort()
    vi_list = [values1, values2]
    values_list = []
    for i, vals in enumerate(vi_list):
        idxs = np.nonzero((all_idxs % 2) == i)[0]
        all_values = vals
        value_diffs = np.diff(all_values)
        values = np.zeros(all_idxs.shape)
        values[idxs[1:]] = value_diffs
        values[idxs[0]] = all_values[0]
        values_list.append(values.cumsum())
    values = np.maximum(values_list[0], values_list[1])
    idxs = all_idxs // 2
    empty_ends = np.nonzero(np.diff(idxs) == 0)[0]
    max_values = np.maximum(values[empty_ends], values[empty_ends+1])
    values[empty_ends+1] = max_values
    values[empty_ends] = max_values

    indices = idxs
    values = values

    indices, values = sanitize_indices_and_values(indices, values)
    return indices, values


def continuous_sparse_maximum(indices1, values1, indices2, values2):
    all_indices = np.concatenate([indices1, indices2])
    # all_values = np.concatenate([values1, values2])
    codes = np.concatenate([np.zeros_like(indices1), np.ones_like(indices2)])
    sorted_args = np.argsort(all_indices)

    sorted_indices = all_indices[sorted_args]
    sorted_codes = codes[sorted_args]
    values_list = []
    for code, values in enumerate(values1, values2):
        my_args = sorted_codes == code
        diffs = np.diff(values)
        my_values = np.zeros(sorted_indices.shape)
        my_values[my_args[1:]] = diffs
        my_values[my_args[0]] = values[0]
        values_list.append(my_values.cumsum())

    values = np.maximum(values_list[0], values_list[1])
    empty_ends = np.nonzero(np.diff(sorted_indices) == 0)[0]
    max_values = np.maximum(values[empty_ends], values[empty_ends+1])
    values[empty_ends+1] = max_values
    values[empty_ends] = max_values
    indices, values = sanitize_indices_and_values(sorted_indices, values)
    return indices, values


def sanitize_indices_and_values(indices, values):

    new_indices = []
    new_values = []

    prev = None
    for i, index in enumerate(indices):
        if values[i] != prev:
            new_indices.append(index)
            new_values.append(values[i])
        prev = values[i]

    return new_indices, new_values


def create_linear_map(ob_graph, snarl_file_name = "haplo1kg50-mhc.snarls", out_file_name="linear_map", copy_graph=True):
    if copy_graph:
        ob_graph = ob_graph.copy()
    builder = SnarlGraphBuilder.from_vg_snarls(
        ob_graph,
        snarl_file_name)
    snarlgraph = builder.build_snarl_graphs()
    linear_map = LinearSnarlMap.from_snarl_graph(snarlgraph, ob_graph)
    linear_map.to_json_files(out_file_name)
    logging.info("Created linear snarl map, wrote to file %s" % out_file_name)

def create_ob_graph_from_vg(vg_json_graph_file_name, ob_graph_file_name="graph.obg"):
    vg_graph = pyvg.Graph.create_from_file(vg_json_graph_file_name)
    ob_graph = vg_graph.get_offset_based_graph()
    ob_graph.to_file(ob_graph_file_name)
    logging.info("Wrote obgraph to %s" % ob_graph_file_name)


def create_linear_path(ob_graph, vg_graph, path_name="ref", write_to_file="linear_path.intervalcollection"):
    assert ob_graph is not None
    linear_paths = get_linear_paths_in_graph(ob_graph, vg_graph, write_to_file_name=write_to_file)
    #ref_path = linear_paths[path_name].to_indexed_interval()
    ref_path = obg.NumpyIndexedInterval.from_interval(linear_paths[path_name])
    return ref_path



import numpy as np
import logging

from ..sparsediffs import SparseDiffs
from .linearsnarls import LinearPileup
from .snarlmaps import LinearSnarlMap


class SparseControl:
    def __init__(self, linear_map, graph, extension_sizes, fragment_length, touched_nodes):
        self._linear_map = LinearSnarlMap.from_json_files(linear_map, graph)
        self._extension_sizes = extension_sizes
        self._fragment_length = fragment_length
        self._graph = graph
        self._min_value = None
        self._touched_nodes = touched_nodes

    def set_min_value(self, value):
        self._min_value = value

    def create(self, reads):
        mapped_reads = self._linear_map.map_interval_collection(reads)
        if self._min_value is None:
            self._min_value = mapped_reads.n_intervals*self._fragment_length / self._linear_map._length
        logging.info("Using min value %s", self._min_value)
        max_pileup = None
        # SparseDiffs([0], [self._min_value])
        for tmp_extension in self._extension_sizes:
            extension = tmp_extension // 2
            sparse_diffs = SparseDiffs.from_starts_and_ends(
                mapped_reads.extend_np(extension))
            sparse_diffs /= (extension*2/self._fragment_length)
            if max_pileup is None:
                sparse_diffs.clip_min(self._min_value)
                max_pileup = sparse_diffs
            else:
                max_pileup = max_pileup.maximum(sparse_diffs)
        max_pileup._sanitize()
        lin_pileup = LinearPileup(
            max_pileup._indices,
            np.cumsum(max_pileup._diffs))
        lin_pileup.sanitize_indices()
        lin_pileup.sanitize_values()
        return lin_pileup.to_sparse_pileup(
            self._linear_map, self._touched_nodes, self._min_value)



import numpy as np
import logging
from collections import defaultdict

from ..eventsorter import EventSorter, EventSort
from .snarlmaps import LinearSnarlMap


def create_control(linear_map_name, *args, **kwargs):
    logging.info("Reading linear map from file")
    linear_map = LinearSnarlMap.from_json_files(linear_map_name, kwargs["ob_graph"])
    return create_control_from_objs(linear_map, *args, **kwargs)


def create_control_from_objs(linear_map, reads, extension_sizes,
                             fragment_length, ob_graph=None, touched_nodes=None,
                             use_global_min_value=None):
    """
    :param snarl_graph: Hierarchical snarl graph
    """
    linear_size = linear_map._length
    mapped_reads = linear_map.map_interval_collection(reads)
    average_value = mapped_reads.n_intervals*fragment_length / linear_size
    logging.info(
        "Average control value: %.4f (sum of pileup: %d, linear size: %d)" % (
            average_value, mapped_reads.n_basepairs_covered(), linear_size))

    if use_global_min_value is not None:
        average_value = use_global_min_value
        logging.warning("Using global min value %.5f" % use_global_min_value)

    max_pileup = LinearPileup([0], [average_value])
    logging.info("Extending control reads with extension sizes: %s" %
                 extension_sizes)
    for tmp_extension in extension_sizes:
        logging.info("Extension size: %d" % tmp_extension)
        extension = tmp_extension // 2
        extended_reads = mapped_reads.extend(extension)
        linear_pileup = LinearPileup.create_from_starts_and_ends(
                extended_reads.starts, extended_reads.ends)
        assert isinstance(linear_pileup, LinearPileup)
        linear_pileup /= (extension*2/fragment_length)
        logging.info("Linear pileup created. Doing maximum")
        max_pileup.maximum(linear_pileup)

    logging.info("All extensions done. Grating valued indexes from pileup")

    logging.info("Making sparsepilup from valued indexes")
    graph_pileup = max_pileup.to_dense_pileup(linear_map, touched_nodes=touched_nodes)
    logging.info("Control pileup created")

    return graph_pileup


class UnmappedIndices(object):
    def __init__(self, indices=None, values=None):
        self.indices = [] if indices is None else indices
        self.values = [] if values is None else values

    def __str__(self):
        return "(%s, %s)" % (self.indices, self.values)

    def get_index_array(self):
        return np.array(self.indices)

    def get_values_array(self):
        return np.array(self.values)

    def add_indexvalue(self, index, value):
        self.indices.append(index)
        self.values.append(value)


class LinearPileup(object):
    def __init__(self, indices, values):
        self.indices = indices
        self.values = values

    def __eq__(self, other):
        if not np.allclose(self.indices, other.indices):
            return False
        return np.allclose(self.values, other.values)

    def __itruediv__(self, scalar):
        self.values /= scalar
        return self

    def __str__(self):
        i = np.array(self.indices)
        v = np.array(self.values)
        pos = i >= 0
        return "Indices: %s, values: %s" % (i[pos],
                                            v[pos])

    __repr__ = __str__

    @classmethod
    def create_from_starts_and_ends(cls, starts, ends):
        logging.info("Creating linear pileup from starts and ends.")
        es = EventSort([starts, ends], [1, -1])
        return LinearPileup(es.indices, es.values)

    def to_sparse_pileup(self, linear_map, touched_nodes=None, min_value=0):
        logging.info("Getting event sorter")
        event_sorter = self.get_event_sorter(linear_map, touched_nodes)
        logging.info("Getting unmapped indices")
        unmapped_indices = self.from_event_sorter(event_sorter)
        logging.info("Mapping linear map to graph pileup")
        #return linear_map.to_numpy_sparse_pileup(unmapped_indices)
        return linear_map.to_sparse_pileup(unmapped_indices, min_value)

    def to_dense_pileup(self, linear_map, touched_nodes=None):
        logging.info("Getting event sorter")
        event_sorter = self.get_event_sorter(linear_map, touched_nodes)
        logging.info("Getting unmapped indices")
        unmapped_indices = self.from_event_sorter(event_sorter)
        logging.info("Mapping linear map to graph pileup")
        return linear_map.to_dense_pileup(unmapped_indices)


    def to_valued_indexes(self, linear_map, touched_nodes=None):
        logging.info("Getting event sorter")
        event_sorter = self.get_event_sorter(linear_map, touched_nodes)
        logging.info("Getting unmapped indices")
        unmapped_indices = self.from_event_sorter(event_sorter)
        logging.info("Mapping linear map to graph pileup")
        vi_dict = linear_map.to_graph_pileup(unmapped_indices)
        return vi_dict

    def get_event_sorter(self, linear_map, touched_nodes=None):
        node_start_values = [node_id for node_id in (linear_map._graph.blocks if touched_nodes is None else touched_nodes)]
        node_end_values = node_start_values[:]
        node_starts_idxs = [linear_map.get_node_start(node_id)
                            for node_id in node_start_values]
        node_end_idxs = [linear_map.get_node_end(node_id)
                         for node_id in node_end_values]
        for start_idx, end_idx in zip(node_starts_idxs, node_end_idxs):
            assert start_idx < end_idx
        idxs = [node_end_idxs, self.indices, node_starts_idxs]
        values = [node_end_values, self.values, node_start_values]
        event_sorter = EventSorter(idxs, values, names=["NODE_END",
                                                        "PILEUP_CHANGE",
                                                        "NODE_START",
                                                        ])
        return event_sorter

    @staticmethod
    def from_event_sorter(event_sorter):
        unmapped_indices = defaultdict(UnmappedIndices)
        cur_nodes = set([])
        cur_index = 0
        cur_value = 0
        for index, code, value in event_sorter:
            if code == event_sorter.NODE_START:
                value = int(value)
                cur_nodes.add(value)
                unmapped_indices[value].add_indexvalue(cur_index, cur_value)
            elif code == event_sorter.PILEUP_CHANGE:
                [unmapped_indices[node_id].add_indexvalue(index, value)
                 for node_id in cur_nodes]
                cur_value = value
                cur_index = index
            elif code == event_sorter.NODE_END:
                cur_nodes.remove(int(value))
            else:
                raise Exception("Coding Error")
        return unmapped_indices

    def continuous_sparse_maximum(self, other):
        indices1 = self.indices
        indices2 = other.indices
        values1 = self.values
        values2 = other.values
        all_indices = np.concatenate([indices1, indices2])
        codes = np.concatenate([np.zeros_like(indices1),
                                np.ones_like(indices2)])
        sorted_args = np.argsort(all_indices)
        sorted_indices = all_indices[sorted_args]
        sorted_codes = codes[sorted_args]
        values_list = []
        for code, values in enumerate((values1, values2)):
            my_args = np.where(sorted_codes == code)[0]
            diffs = np.diff(values)
            my_values = np.zeros(sorted_indices.shape)
            my_values[my_args[1:]] = diffs
            my_values[my_args[0]] = values[0]
            values_list.append(my_values.cumsum())
        values = np.maximum(values_list[0], values_list[1])
        self.indices = sorted_indices
        self.values = values
        self.sanitize_indices()
        self.sanitize_values()
        # 
        # 
        # empty_ends = np.nonzero(np.diff(sorted_indices) == 0)[0]
        # max_values = np.maximum(values[empty_ends], values[empty_ends+1])
        # values[empty_ends+1] = max_values
        # values[empty_ends] = max_values
        # indices, values = sanitize_indices_and_values(sorted_indices, values)
        # self.indices = indices
        # self.values = values

    def sanitize_indices(self, choose_last=True):
        assert choose_last
        idx_diffs = np.diff(self.indices)
        changes = np.nonzero(idx_diffs)[0]
        new_args = np.concatenate([changes, [self.values.size - 1]])
        self.indices = self.indices[new_args]
        self.values = self.values[new_args]

    def sanitize_values(self):
        value_diffs = np.diff(self.values)
        changes = np.nonzero(value_diffs)[0]
        new_args = np.concatenate([[0], changes+1])
        self.indices = self.indices[new_args]
        self.values = self.values[new_args]

    def maximum(self, other):
        return self.continuous_sparse_maximum(other)



import pickle


class ExperimentInfo(object):
    def __init__(self, genome_size, fragment_length, read_length):
        self.genome_size = genome_size
        self.n_sample_reads = 0  # Counters will be modified by Callpeaks
        self.n_control_reads = 0
        self.fragment_length = fragment_length
        self.read_length = read_length

    def to_file(self, file_name):
        with open("%s" % file_name, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def from_file(cls, file_name):
        with open("%s" % file_name, "rb") as f:
            data = pickle.loads(f.read())
            return data



from .extender import Areas
import numpy as np
import pickle
import logging
from collections import OrderedDict
from .areas import BinaryContinousAreas
from .extender import Areas
import itertools
from collections import defaultdict

class ConnectedAreas(Areas):
    def __init__(self, graph, areas=None):
        super(ConnectedAreas, self).__init__(graph, areas)

    def touches_area(self, other_node, start, end):
        graph = self.graph

        if start > 0 and end < graph.node_size(other_node):
            if other_node not in self.areas:
                return False

        if start == 0:
            if not self._is_start_position(other_node, start):
                return True

        if end == self.graph.node_size(other_node):
            if not self._is_end_position(other_node, end):
                return True

        return False

    def __add__(self, other):
        # Adds another connected area to this one
        for node_id, starts_and_ends in other.areas.items():
            self.add_areas_for_node(node_id, starts_and_ends)
        return self

    def contains_interval(self, interval):
        intervals = self.to_simple_intervals()
        overlap = 0
        for internal_interval in intervals:
            overlap += interval.overlap(internal_interval)

        assert overlap <= interval.length()

        if overlap == interval.length():
            return True

        return False

    def n_basepairs(self):
        intervals = self.to_simple_intervals()
        n = 0
        for interval in intervals:
            n += interval.length()
        return n


class BCConnectedAreas(BinaryContinousAreas):
    def __init__(self, graph, id, areas=None):
        super(BCConnectedAreas, self).__init__(graph)
        self.id = id
        for node, startend in areas.items():
            self.add(node, startend[0], startend[1])

    def __hash__(self):
        return self.id

    def touches_area(self, other_node, start, end):
        graph = self.graph

        if start > 0 and end < graph.node_size(other_node):
            if other_node not in self.areas:
                return False

        if start == 0:
            if not self._is_start_position(other_node, start):
                return True

        if end == self.graph.node_size(other_node):
            if not self._is_end_position(other_node, end):
                return True

        return False

    def __add__(self, other):
        # Adds another connected area to this one
        self.merge_with_other(other)
        return self

    def contains_interval(self, interval):
        intervals = self.to_simple_intervals()
        overlap = 0
        for internal_interval in intervals:
            overlap += interval.overlap(internal_interval)

        assert overlap <= interval.length()

        if overlap == interval.length():
            return True

        return False

    def n_basepairs(self):
        intervals = self.to_simple_intervals()
        n = 0
        for interval in intervals:
            n += interval.length()
        return n


class SubgraphCollection(object):

    def __init__(self, graph, subgraphs=None):
        self.graph = graph
        self._id_counter = 0
        if subgraphs is not None:
            self.subgraphs = subgraphs
        else:
            self.subgraphs = []

        self._node_end_index = defaultdict(set)  # Index from node id to subgraphs touching end
        self._node_start_index = defaultdict(set)

    def remove(self, subgraph):
        self.subgraphs.remove(subgraph)

    def add(self, subgraph):
        pass

    @classmethod
    def from_pileup(cls, graph, pileup):
        collection = cls(graph)
        logging.info("Finding valued areas")
        areas = pileup.find_valued_areas(1)
        n = 0
        for node_id, starts_and_ends in areas.items():
            if n % 50000 == 0:
                logging.info("Adding node %d to subgraph" % n)
            n += 1
            for i in range(0, len(starts_and_ends) // 2):
                collection.add_area(node_id,
                                    starts_and_ends[i*2],
                                    starts_and_ends[i*2+1])
        return collection

    def _subgraphs_touching_area(self, node_id, start, end):
        out = set()
        if start == 0:
            #print("   Checking start of node %d" % node_id)
            out = self._node_start_index[node_id]
            #print("   Found:")
            #print(self._node_start_index[node_id])
        if end == self.graph.node_size(node_id):
            out.update(self._node_end_index[node_id])
            #print("  Checking end of node %d" % node_id)
            #print("   Found:")
            #print(self._node_end_index[node_id])
        #print("   Found in total %d touching subgraphs" % len(out))
        #return list(set(out))
        return out

        touching_subgraphs = []
        for subgraph in self.subgraphs:
            if subgraph.touches_area(node_id, start, end):
                touching_subgraphs.append(subgraph)
        return touching_subgraphs

    def __iter__(self):
        return iter(self.subgraphs)

    def add_indexes(self, node, start, end, subgraph):
        # Checks if added node id,start,end touches nodes. Link index to subgraph
        if start == 0:
            for in_node in self.graph.adj_list[-node]:
                #print("    Adding subgraph to end index of %d and start of %d" % (in_node, -in_node))
                self._node_end_index[in_node].add(subgraph)
                self._node_start_index[-in_node].add(subgraph)
            for in_node in self.graph.reverse_adj_list[-node]:
                #print("    Adding subgraph to end index of %d and end of %d" % (-in_node, in_node))
                self._node_end_index[-in_node].add(subgraph)
                self._node_start_index[in_node].add(subgraph)

        if end == self.graph.node_size(node):
            for in_node in self.graph.adj_list[node]:
                #print("    Adding subgraph to start index of %d and end index of %d" % (in_node, -in_node))
                self._node_start_index[in_node].add(subgraph)
                self._node_end_index[-in_node].add(subgraph)
            for in_node in self.graph.reverse_adj_list[node]:
                #print("    Adding subgraph to end index of %d and start of %d" % (-in_node, in_node))
                self._node_end_index[-in_node].add(subgraph)
                self._node_start_index[in_node].add(subgraph)



    def add_area(self, node_id, start, end):
        assert node_id in self.graph.blocks
        assert start >= 0 and start < end
        assert end <= self.graph.node_size(node_id)
        #print("Adding node %d, startend %d %d" % (node_id, start, end))
        #print("%d subgraphs so far" % len(self.subgraphs))
        if start > 0 and end < self.graph.node_size(node_id):
            # Internal, will never touch anything
            touching_subgraphs = []
        else:
            touching_subgraphs = list(self._subgraphs_touching_area(node_id, start, end))

        self._id_counter += 1
        if len(touching_subgraphs) == 0:
            new_subgraph = BCConnectedAreas(
                self.graph,
                self._id_counter,
                {node_id: np.array([start, end])})
            self.subgraphs.append(new_subgraph)
            self.add_indexes(node_id, start, end, new_subgraph)

        elif len(touching_subgraphs) == 1:
            touching_subgraphs[0].add(
                node_id, start, end)
            #print("Subgraph after adding")
            #print(touching_subgraphs[0])
            self.add_indexes(node_id, start, end, touching_subgraphs[0])

        elif len(touching_subgraphs) > 1:
            #print("  Found multiple subgraphs")
            # Merge all touching subgraphs, then add the area
            new_subgraph = touching_subgraphs[0]
            for touching_subgraph in touching_subgraphs[1:]:
                new_subgraph = new_subgraph + touching_subgraph
                if touching_subgraph in self.subgraphs:
                    self.remove(touching_subgraph)

            new_subgraph.add(node_id, start, end)
            self.add_indexes(node_id, start, end, new_subgraph)

    def to_file(self, file_name):
        f = open(file_name, "w")
        lines = (subgraph.to_file_line() for subgraph in self.subgraphs)
        f.writelines(lines)
        f.close()

    def contains_interval(self, interval):
        for connected_area in self.subgraphs:
            if connected_area.contains_interval(interval):
                return True

        return False

    @classmethod
    def from_pickle(cls, file_name, graph=None):
        with open("%s" % file_name, "rb") as f:
            obj = pickle.loads(f.read())
            assert isinstance(obj, cls)

            for subgraph in obj.subgraphs:
                subgraph.graph = graph

            return obj

    def to_pickle(self, file_name):
        with open("%s" % file_name, "wb") as f:
            pickle.dump(self, f)


class SubgraphCollectionPartiallyOrderedGraph(SubgraphCollection):
    @classmethod
    def create_from_pileup(cls, graph, pileup):
        builder = SubgraphCollectionPartiallyOrderedGraphBuilder(
                    graph, pileup)
        return builder.build()


class SingleArea:
    def __init__(self, node, start, end, is_start, is_end):
        self.node = node
        self.start = start
        self.end = end
        self.is_start = is_start
        self.is_end = is_end

    def __str__(self):
        return "Node %d, start/end: %d/%d. Is start: %s, is end:%s" % \
               (self.node, self.start, self.end, self.is_start, self.is_end)

    def __repr__(self):
        return self.__str__()


class SubgraphCollectionPartiallyOrderedGraphBuilder():

    def __init__(self, graph, pileup):
        self.graph = graph
        self.pileup = pileup
        from .densepileup import DensePileupData
        assert isinstance(self.pileup.data, DensePileupData)
        areas = pileup.find_valued_areas(1)
        self.areas = Areas(self.graph, areas)
        self.single_areas = []

    def _make_single_areas(self):
        logging.info("Making single areas")
        sorted_nodes = sorted(self.areas.areas.keys())
        n = 0
        for node_id in sorted_nodes: #, starts_and_ends in self.areas.areas.items():
            #print("Checking node %d" % node_id)
            starts_and_ends = self.areas.areas[node_id]
            if n % 1000000 == 0:
                logging.info("Adding node %d to subgraph" % n)
            n += 1
            for i in range(0, len(starts_and_ends) // 2):
                start = starts_and_ends[i*2]
                end = starts_and_ends[i*2+1]
                is_start = self.areas._is_start_position(node_id, start)
                #print("Checking end %d, %d" % (node_id, end))
                is_end = self.areas._is_end_position(node_id, end)
                yield SingleArea(node_id, start, end, is_start, is_end)

    def _get_connected_subgraphs(self, subgraphs, area):
        connected = []
        #print("Getting connected subgraphs")
        for subgraph in subgraphs:
            #print("  Checking subgraph")
            if not area.is_start:
                for prev_node in self.graph.reverse_adj_list[-area.node]:
                    #print("        Checking prev node %d" % prev_node)
                    if -prev_node in subgraph.full_areas:
                        connected.append(subgraph)
                    if prev_node in subgraph.starts:
                        connected.append(subgraph)
            elif not area.is_end:
                for next_node in self.graph.adj_list[area.node]:
                    #print("        Checking next node %d" % next_node)
                    if next_node in subgraph.starts:
                        connected.append(subgraph)
                    if next_node in subgraph.full_areas:
                        connected.append(subgraph)

        return connected

    def _create_subgraphs(self):
        from collections import deque
        active_subgraphs = deque()

        logging.info("Creating subgraphs after single areas are created")
        peaks = []
        current_peak = BinaryContinousAreas(self.graph)
        prev_was_end = False
        prev_node = 0
        for single_area in self.single_areas:
            #print("Single area: %s " % single_area)

            connected = self._get_connected_subgraphs(active_subgraphs, single_area)
            if len(connected) > 1:
                merged = connected[0]
                for subgraph in connected[1:]:
                    merged.merge_with_other(subgraph)

                connected = [merged]

            if len(connected) == 0:
                current_peak = BinaryContinousAreas(self.graph)
                current_peak.add(single_area.node, single_area.start,
                                 single_area.end)
                if single_area.is_start and single_area.is_end:
                    peaks.append(current_peak)
                    #print("   Creating new, finishing it")
                else:
                    active_subgraphs.append(current_peak)
                    #print("   Creating new, appending to active")
            else:
                #print("   Adding to active")
                current_peak = connected[0]

                current_peak.add(single_area.node,
                             single_area.start,
                             single_area.end)

            # If more than 2, the first must be finished
            if len(active_subgraphs) > 5:
                peaks.append(active_subgraphs.popleft())
                #print("Finishing subgraph")

        peaks.extend(active_subgraphs)

        return peaks


    def build(self):
        self.single_areas = self._make_single_areas()
        return self._create_subgraphs()

    """
    def build(self):
        sorted_nodes = sorted(self.areas.keys())

        starts = OrderedDict()
        ends = OrderedDict()

        for node in sorted_nodes:
            starts_ends = self.areas[node]
            assert len(starts_ends) == 2 or len(starts_ends) == 4

            if self.areas._is_start_position(node, starts_ends[0]):
                starts[node] = starts_ends[:2]

            if self.areas._is_end_position(node, starts_ends[-1]):
                ends[node] = starts_ends[-2:]


        # Iterate sorted nodes. Count number of starts and number of ends
        in_peak = False
        prev_status = in_peak
        peaks = []
        current_peak = BinaryContinousAreas(self.graph)
        for node in sorted_nodes:
            start_ends = self.areas[node]
            assert len(start_ends) == 2 or len(starts_ends) == 4
            is_start = node in starts
            is_end = node in ends

            if is_start and is_end:
                current_peak.add(node, start_ends[0:2])
    """



from offsetbasedgraph import IntervalCollection
from collections import defaultdict
from offsetbasedgraph import IndexedInterval
import logging

class HaploTyper:
    """
    Simple naive haplotyper
    """
    def __init__(self, graph, intervals):
        assert isinstance(intervals, IntervalCollection)
        self.graph = graph
        self.intervals = intervals

        self.node_counts = defaultdict(int)

    def build(self):
        logging.info("Building haplotyper")
        i = 0
        for interval in self.intervals.intervals:
            if i % 50000 == 0:
                logging.info("%d intervals processed" % i)
            i += 1

            for rp in interval.region_paths:
                self.node_counts[rp] += 1

    def get_maximum_interval_through_graph(self):
        logging.info("Getting first blocks")
        graph = self.graph
        start = graph.get_first_blocks()
        logging.info("First blocks found")
        assert len(start) == 1, "Only works when graph has one start node"
        nodes = []
        current_block = start[0]
        i = 0
        while True:
            if i % 1000000 == 0:
                logging.info("Processing node %d" % i)
            i += 1
            nodes.append(current_block)
            next_blocks = graph.adj_list[current_block]
            if len(next_blocks) < 1:
                break

            next = None
            max_pileup_value = -1
            for potential_next in next_blocks:
                pileup_value = self.node_counts[potential_next]
                if pileup_value > max_pileup_value:
                    next = potential_next
                    max_pileup_value = pileup_value

            current_block = next

        return IndexedInterval(0, graph.node_size(nodes[-1]), nodes, graph)



import logging
import pickle

import offsetbasedgraph as obg
from .sparsepvalues import PValuesFinder, PToQValuesMapper
from .experiment_info import ExperimentInfo
from .control import SparseControl
from .sample import SamplePileupGenerator


obg.IntervalCollection.interval_class = obg.DirectedInterval


def enable_filewrite(func):
    def wrapper(*args, **kwargs):
        intervals = args[1]
        if isinstance(intervals, str):
            intervals = obg.IntervalCollection.from_file(intervals)

        write_to_file = kwargs.pop("write_to_file", False)
        interval_list = func(args[0], intervals, **kwargs)

        if write_to_file:
            interval_collection = obg.IntervalCollection(interval_list)
            interval_collection.to_file(write_to_file)
            return write_to_file
        else:
            return interval_list

    return wrapper


class SampleAndControlCreator:
    def __init__(self, graph, sample_intervals,
                 control_intervals=None, experiment_info=None,
                 verbose=False, out_file_base_name="", has_control=True,
                 linear_map=None, skip_filter_duplicates=False,
                 skip_read_validation=False,
                 save_tmp_results_to_file=True,
                 configuration=None,
                 graph_is_partially_ordered=False):
        """
        :param sample_intervals: Either an interval collection or file namen
        :param control_intervals: Either an interval collection or a file name
        """

        assert linear_map is not None, "LinearMap cannot be None"
        assert isinstance(linear_map, str), "Must be file name"

        assert isinstance(sample_intervals, obg.IntervalCollection) \
               or isinstance(sample_intervals, str), \
                "Samples intervals must be either interval collection or a file name"

        assert isinstance(control_intervals, obg.IntervalCollection) \
               or isinstance(control_intervals, str), \
                "control_intervals must be either interval collection or a file name"

        if not has_control:
            logging.info("Running without control")
        else:
            logging.info("Running with control")

        self.graph = graph

        self.sample_intervals = sample_intervals
        self.control_intervals = control_intervals
        self.has_control = has_control
        self.linear_map = linear_map

        self._p_value_track = "p_value_track"
        self._q_value_track = "q_value_track"
        self.info = experiment_info
        self.verbose = verbose
        self._control_pileup = None
        self._sample_pileup = None
        self.out_file_base_name = out_file_base_name
        self.pre_processed_peaks = None
        self.filtered_peaks = None
        self.skip_filter_duplicates = skip_filter_duplicates
        self.skip_read_validation = skip_read_validation
        self.save_tmp_results_to_file = save_tmp_results_to_file
        self.n_duplicates_found_sample = 0

        self.max_paths = None
        self.peaks_as_subgraphs = None

        self.create_graph()
        self.touched_nodes = None  # Nodes touched by sample pileup
        self.graph_is_partially_ordered = graph_is_partially_ordered

        # Tmp hack, shuold be default
        if configuration is not None:
            self.graph_is_partially_ordered = configuration.graph_is_partially_ordered
            self.skip_filter_duplicates = configuration.skip_filter_duplicates
            self.skip_read_validation = configuration.skip_read_validation
            self.save_tmp_results_to_file = configuration.save_tmp_results_to_file
            self.use_global_min_value = configuration.use_global_min_value


            if self.use_global_min_value is not None:
                logging.info("Will use global min value %.4f when creating background signal" % self.use_global_min_value)

        if self.skip_filter_duplicates:
            logging.warning("Not removing duplicates")
        else:
            logging.info("Will remove duplicates.")

    def run(self):
        self.preprocess()
        if self.info is None:
            self.info = ExperimentInfo.find_info(
                self.ob_graph, self.sample_intervals, self.control_intervals)
        self.create_sample_pileup()
        self.create_control()
        self.scale_tracks()

    def preprocess(self):
        self.info.n_control_reads = 0
        self.info.n_sample_reads = 0

        if not self.skip_read_validation:
            self.sample_intervals = self.remove_alignments_not_in_graph(
                                        self.sample_intervals)
        else:
            logging.warning("Skipping validation of reads. Not checking whether reads are valid"
                            " or inside the graph.")

        self.sample_intervals = self.filter_duplicates_and_count_intervals(
                                    self.sample_intervals, is_control=False)

        if not self.skip_read_validation:
            self.control_intervals = self.remove_alignments_not_in_graph(
                                    self.control_intervals, is_control=True)
        self.control_intervals = self.filter_duplicates_and_count_intervals(
                                    self.control_intervals, is_control=True)

    @enable_filewrite
    def remove_alignments_not_in_graph(self, intervals, is_control=False):
        for interval in self._get_intervals_in_ob_graph(intervals):
            if interval is not False:
                yield interval

    @enable_filewrite
    def filter_duplicates_and_count_intervals(self, intervals, is_control=False):
        interval_hashes = {}
        n_duplicates = 0
        for interval in intervals:
            if not self.skip_filter_duplicates:
                hash = interval.hash(ignore_end_pos=True)
                if hash in interval_hashes:
                    n_duplicates += 1
                    if not is_control:
                        self.n_duplicates_found_sample += 1
                    continue

                interval_hashes[hash] = True

            if is_control:
                self.info.n_control_reads += 1
            else:
                self.info.n_sample_reads += 1

            yield interval

    def _assert_interval_is_valid(self, interval):
        # Checks that the interval (read) is a valid connected interval
        direction = None
        for i, rp in enumerate(interval.region_paths[:-1]):
            next_rp = interval.region_paths[i+1]
            if next_rp in self.ob_graph.adj_list[rp]:
                new_dir = 1
            elif next_rp in self.ob_graph.reverse_adj_list[rp]:
                new_dir = -1
            else:
                logging.error("Invalid interval: Rp %d of interval %s is not "
                              "connected in graph to rp %d, which is the next rp"
                              % (rp, interval, next_rp))
                raise Exception("Invalid interval")

            if direction is None:
                direction = new_dir
            else:
                if new_dir != direction:
                    logging.error("Invalid read: Interval %s is going edges in multiple directions.")
                    raise Exception("Invalid interval")

        return True

    def _get_intervals_in_ob_graph(self, intervals):
        # Returns only those intervals that exist in graph
        for interval in intervals:
            self._assert_interval_is_valid(interval)
            if interval.region_paths[0] in self.ob_graph.blocks:
                yield interval
            else:
                logging.warning("Interval: %s" % interval)
                raise Exception("Interval not in graph")

    def scale_tracks(self, update_saved_files=False):
        logging.info("Scaling tracks to ratio: %d / %d" % (
            self.info.n_sample_reads,
            self.info.n_control_reads))
        ratio = self.info.n_sample_reads/self.info.n_control_reads

        if self.info.n_sample_reads == self.info.n_control_reads:
            logging.info("Not scaling any tracks because of same amount of reads")
            self._control_pileup.to_bed_graph(
                self.out_file_base_name + "scaled_control.bdg")
            return

        if ratio > 1:
            logging.warning("More reads in sample than in control")
            self._sample_pileup *= ratio
            self._sample_pileup.scale(1/ratio)
            self._sample_pileup.to_bed_graph(
                self.out_file_base_name + "scaled_treat.bdg")
            if update_saved_files:
                self._sample_pileup.to_bed_graph(self._sample_track)
        else:
            logging.info("Scaling control pileup down using ration %.3f" % ratio)
            self._control_pileup *= ratio  # .scale(ratio)
            self._control_pileup.to_bed_graph(
                self.out_file_base_name + "scaled_control.bdg")
            if update_saved_files:
                self._control_pileup.to_bed_graph(self._control_track)

    def find_info(self):
        sizes = (block.length() for block in self.ob_graph.blocks.values())

        self.genome_size = sum(sizes)
        self.n_reads = sum(1 for line in open(self.control_file_name))

    def create_graph(self):
        logging.info("Creating graph")
        if isinstance(self.graph, str):
            self.ob_graph = obg.Graph.from_file(self.graph)
        else:
            self.ob_graph = self.graph
            logging.info("Graph already created")

    def create_control(self):
        logging.info("Creating control track using linear map %s" % self.linear_map)
        extensions = [self.info.fragment_length, 1000, 10000] if self.has_control else [10000]
        sparse_control = SparseControl(
            self.linear_map, self.graph, extensions,
            self.info.fragment_length, self.touched_nodes)
        if self.use_global_min_value:
            sparse_control.set_min_value(self.use_global_min_value)
        control_pileup = sparse_control.create(self.control_intervals)
        control_pileup.to_sparse_files(
            self.out_file_base_name + "control_track")
        self.linear_map = None

        # control_pileup.graph = self.ob_graph
        logging.info("Number of control reads: %d" % self.info.n_control_reads)

        if self.save_tmp_results_to_file and False:
            self._control_track = self.out_file_base_name + "control_track.bdg"
            control_pileup.to_bed_graph(self._control_track)
            logging.info("Saved control pileup to " + self._control_track)

        self._control_pileup = control_pileup

        # Delete control pileup
        self.control_intervals = None

    def get_score(self):
        logging.info("Getting p valyes.")
        p_values_finder = PValuesFinder(
            self._sample_pileup, self._control_pileup)
        self._p_values_pileup = p_values_finder.get_p_values_pileup()

        # Delete sample and control pileups
        self._control_pileup = None
        self._sample_pileup = None

    def get_p_to_q_values_mapping(self):
        return PToQValuesMapper.from_p_values_dense_pileup(self.p_values)

    def _write_vg_alignments_as_intervals_to_bed_file(self):
        pass

    def create_sample_pileup(self):
        logging.debug("In sample pileup")
        logging.info("Creating sample pileup")
        logging.info(self.sample_intervals)
        logging.info("Processing areas")
        generator = SamplePileupGenerator(
            self.graph, self.info.fragment_length-self.info.read_length)
        pileup = generator.run(self.sample_intervals,
                               self.out_file_base_name+"direct_pileup")
        self.touched_nodes = pileup.touched_nodes

        self._sample_track = self.out_file_base_name + "sample_track"
        if self.save_tmp_results_to_file or True:
            logging.info("Saving sample pileup to file")
            pileup.to_sparse_files(self._sample_track)
            logging.info("Saved sample pileup to " + self._sample_track)

            logging.info("Writing touched nodes to file")
            with open(self.out_file_base_name + "touched_nodes.pickle", "wb") as f:
                pickle.dump(self.touched_nodes, f)

            logging.info("N touched nodes: %d" % len(self.touched_nodes))
        else:
            logging.info("Not saving sample pileup to files.")

        self._sample_pileup = pileup
        logging.info("Found in total %d duplicate reads that were removed" % self.n_duplicates_found_sample)

        # Delete sample intervals
        self.sample_intervals = None



import pickle
import numpy as np
import os
from scipy.stats import poisson
import logging
import math
from .sparsediffs import SparseValues


class PValuesFinder:
    def __init__(self, sample_pileup, control_pileup):
        self.sample = sample_pileup
        self.control = control_pileup

    def get_p_values_pileup(self):
        baseEtoTen = np.log(10)

        def clean_p_values(counts, lambdas):
            p_values = poisson.logsf(counts, lambdas)
            p_values /= -baseEtoTen
            p_values[counts == 0] = 0
            p_values[np.isinf(p_values)] = 1000
            return p_values

        p_values = self.sample.apply_binary_func(
            clean_p_values, self.control,
            return_values=True)

        return p_values


class PToQValuesMapper:

    def __init__(self, p_values, cum_counts):
        self.p_values = p_values
        self.cum_counts = cum_counts

    @classmethod
    def __read_file(cls, file_name):
        indices = np.load(file_name + "_indexes.npy")
        values = np.load(file_name + "_values.npy")
        return indices, values

    @classmethod
    def _from_subcounts(cls, p_values, counts):
        p_values = p_values.ravel()
        counts = counts.ravel()
        args = np.argsort(p_values)[::-1]
        sorted_ps = p_values[args]
        sorted_lens = counts[args]
        cum_counts = np.cumsum(sorted_lens)
        changes = np.ediff1d(sorted_ps, to_end=1) != 0
        cum_counts = cum_counts[changes]
        return cls(sorted_ps[changes], cum_counts)

    @classmethod
    def from_p_values_pileup(cls, p_values):
        logging.info("Creating mapping from p value dense pileup")
        sub_counts = cls.__get_sub_counts(p_values)
        # sub_counts = np.ediff1d(
        #     p_values.indices,
        #     to_end=p_values.track_size-p_values.indices[-1])
        return cls._from_subcounts(p_values.values, sub_counts)

    @classmethod
    def __get_sub_counts(cls, sparse_values):
        return np.ediff1d(
            sparse_values.indices,
            to_end=sparse_values.track_size-sparse_values.indices[-1])

    @classmethod
    def from_files(cls, base_file_name):
        search = base_file_name
        logging.info("Searching for files starting with %s" % search)
        files = (f for f in os.listdir()
                 if f.startswith(search) and
                 "pvalues" in f and f.endswith("_indexes.npy"))
        sub_counts = []
        p_values = []
        for filename in files:
            base_file_name = filename.replace("_indexes.npy", "")
            logging.info("Reading p values from file %s" % base_file_name)
            chr_p_values = SparseValues.from_sparse_files(base_file_name)
            sub_counts.append(cls.__get_sub_counts(chr_p_values))
            p_values.append(chr_p_values.values)
            assert sub_counts[-1].size == p_values[-1].size

        return cls._from_subcounts(
            np.concatenate(p_values),
            np.concatenate(sub_counts))

    def get_p_to_q_values(self):
        logN = np.log10(self.cum_counts[-1])
        q_values = self.p_values + np.log10(
            1+np.r_[0, self.cum_counts[:-1]])-logN
        q_values[0] = max(0, q_values[0])
        q_values = np.minimum.accumulate(q_values)
        d = dict(zip(self.p_values, q_values))
        d[0] = 0
        return d

    def to_file(self, base_name):
        with open(base_name + 'p2q.pkl', 'wb') as f:
            pickle.dump(self._p_to_q_values, f, pickle.HIGHEST_PROTOCOL)


class QValuesFinder:
    def __init__(self, p_values_pileup, p_to_q_values):
        assert isinstance(p_to_q_values, dict)
        self.p_values = p_values_pileup
        self.p_to_q_values = p_to_q_values

    def get_q_values(self):
        q_values = SparseValues(
            self.p_values.indices,
            self.get_q_array_from_p_array(self.p_values.values))
        return q_values

    def get_q_array_from_p_array(self, p_values):
        assert isinstance(p_values, np.ndarray)

        def translation(x):
            # if abs(x) < 1e-9:
            #    return 0
            # if math.isnan(x):
            #    return 0
            # x = "%.7f" % x
            # if x not in self.p_to_q_values:
            #     print(self.p_to_q_values)
            #     print(x)
            #     logging.error("P value not found in mapping dict. Could be due to rounding errors.")
            return self.p_to_q_values[x]

        trans = np.vectorize(translation, otypes=[np.float])
        new_values = trans(p_values)
        return new_values
